/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * From: @(#)af.c	5.11 (Berkeley) 2/28/91
 * From: @(#)af.c	8.1 (Berkeley) 6/5/93
 * From: @(#)af.c	1.5 (dholland Exp)  1996/11/25
 */

char af_rcsid[] = 
  "$Id: af.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

#include "defs.h"

#ifndef PICA_LINUX
#define IN_EXPERIMENTAL(A) A?1:0
#endif

/*
 * Address family support routines
 */

void inet_canon(struct sockaddr *);
int inet_checkhost(struct sockaddr *);
char *inet_format(struct sockaddr *, char *, size_t);
void inet_hash(struct sockaddr *, struct afhash *);
int inet_netmatch(struct sockaddr *, struct sockaddr *);
int inet_portcheck(struct sockaddr *);
int inet_portmatch(struct sockaddr *);
void inet_output(int, int, struct sockaddr *, int);

#define NIL	{ 0 }
#define	INET \
	{ inet_hash,		inet_netmatch,  	inet_output, \
	  inet_portmatch,	inet_portcheck,  	inet_checkhost, \
	  inet_rtflags,		inet_sendroute, 	inet_canon, \
	  inet_format\
	}

struct afswitch afswitch[AF_MAX] = 
{
	NIL,		/* 0- unused */
	NIL,		/* 1- Unix domain, unused */
	INET,		/* Internet */
};

int af_max = sizeof(afswitch) / sizeof(afswitch[0]);

struct sockaddr_in inet_default = 
{
	AF_INET, INADDR_ANY 
};

void inet_hash(struct sockaddr *sa, struct afhash *hp)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	int n;

	n = inet_netof_subnet(sin->sin_addr);
	if (n)
	    while ((n & 0xff) == 0)
		n >>= 8;
	hp->afh_nethash = n;
	hp->afh_hosthash = ntohl(sin->sin_addr.s_addr);
	hp->afh_hosthash &= 0x7fffffff;
}

int inet_netmatch(struct sockaddr *sa1, struct sockaddr *sa2)
{
	struct sockaddr_in *sin1=(struct sockaddr_in *)sa1;
	struct sockaddr_in *sin2=(struct sockaddr_in *)sa2;
	return (inet_netof_subnet(sin1->sin_addr) ==
	    inet_netof_subnet(sin2->sin_addr));
}


/*
 * Verify the message is from the right port.
 */

int inet_portmatch(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	return (ntohs(sin->sin_port) == ntohs(sp->s_port));
}

/*
 * Verify the message is from a "trusted" port.
 */
int inet_portcheck(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	return (ntohs(sin->sin_port) <= IPPORT_RESERVED);
}

/*
 * Internet output routine. 
 */
void inet_output(int s, int flags, struct sockaddr *sa, int size)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	struct sockaddr_in dst;

	dst = *sin;
	sin = &dst;
	if (sin->sin_port == 0)
		sin->sin_port = sp->s_port;
	if (sendto(s, packet, size, flags,
	    (struct sockaddr *)sin, sizeof (*sin)) < 0)
		perror("sendto");

        outputsize = 0;        
}


/*
 * Return 1 if the address is believed
 * for an Internet host -- THIS IS A KLUDGE.
 */

int inet_checkhost(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	int i = ntohl(sin->sin_addr.s_addr);

	//if (IN_EXPERIMENTAL(i) || sin->sin_port != 0)
	if (sin->sin_port != 0)	
		return (0);
	if (i != 0 && (i & 0xff000000) == 0)
		return (0);
	for (i = 0; i < sizeof(sin->sin_zero)/sizeof(sin->sin_zero[0]); i++)
		if (sin->sin_zero[i])
			return (0);
	return (1);
}

void inet_canon(struct sockaddr *sa)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	sin->sin_port = 0;

#ifdef PICA_LINUX
        bzero(sin->sin_zero, sizeof(sin->sin_zero));
#else
	ZeroMemory((void *) sin->sin_zero, sizeof(sin->sin_zero));
#endif

}

char *inet_format(struct sockaddr *sa, char *buf, size_t sz)
{
	struct sockaddr_in *sin=(struct sockaddr_in *)sa;
	strncpy(buf, inet_ntoa(sin->sin_addr), sz);
	buf[sz - 1] = '\0';
	return (buf);
}
