/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char dupplicate_table_rcsid[] = 
  "$Id: duplicate_table.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


#include "defs.h"

struct duplicatehash              duplicatetable[HASHSIZE];

/*------------------------------------------------------------------------*/
void
olsr_delete_duplicate_table(struct duplicate_entry *dup_entry)
{
  olsr_remque((struct olsr_qelem *)dup_entry);
  free((void *)dup_entry);
}
/*------------------------------------------------------------------------*/
int
olsr_insert_duplicate_table(struct tc_message *message)
{
  olsr_u32_t  hash;
  struct duplicatehash *dup_hash;
  struct duplicate_entry *dup_message;
  
  dup_message=(struct duplicate_entry *)malloc(sizeof(struct duplicate_entry));
  dup_message->duplicate_seq=message->packet_seq_number;
  memcpy(&dup_message->duplicate_addr,&message->originator,sizeof(olsr_ip_addr));  
  timeradd(&now,&hold_time_duplicate_tc,&dup_message->duplicate_timer);
  olsr_hashing(&dup_message->duplicate_addr,&hash);
  dup_message->duplicate_hash=hash;
  dup_hash=&duplicatetable[hash & HASHMASK];
  olsr_insque((struct olsr_qelem *)dup_message, (struct olsr_qelem *)dup_hash);
}

/*------------------------------------------------------------------------*/
struct duplicate_entry
*olsr_lookup_duplicate_table(struct tc_message *message)
{
  struct duplicate_entry *dup_message;
  struct duplicatehash   *dup_hash;
  olsr_u32_t             hash;

  olsr_hashing(&message->originator,&hash);
  
  dup_hash=&duplicatetable[hash & HASHMASK];
  
  for(dup_message=dup_hash->duplicate_forw;dup_message!=(struct duplicate_entry *)dup_hash;dup_message=dup_message->duplicate_forw)
    {
      if(dup_message->duplicate_hash!=hash)
	continue;

      if ((0==memcmp(&dup_message->duplicate_addr,&message->originator,sizeof(olsr_ip_addr)))&&(dup_message->duplicate_seq==message->packet_seq_number))
	{

	  /*------------------*/
	  if(debug)
	    {
	      printf("THIS TC MESSAGE IS IN THE DUPLICATE TABLE YET seq_nu of this message is %d  and in the TC Table is %d\n",message->packet_seq_number,dup_message->duplicate_seq);
	    }
	  /*------------------*/

	  return (dup_message);
	}
    }
  return (NULL);
}

/*------------------------------------------------------------------------*/

int
olsr_release_duplicate_table()
{
  olsr_u8_t              index;
  struct duplicatehash   *dup_hash;
  struct duplicate_entry *dup_message;
  struct duplicate_entry *dup_message_tmp;

  for(index=0;index<HASHSIZE;index++)
    {
      dup_hash=&duplicatetable[index];
      dup_message=dup_hash->duplicate_forw;
      while(dup_message!=(struct duplicate_entry *)dup_hash)
	{
	  dup_message_tmp=dup_message;
	  dup_message=dup_message->duplicate_forw;
	  olsr_delete_duplicate_table(dup_message_tmp);
	}
    }
}
/*------------------------------------------------------------------------*/
int
olsr_time_out_duplicate_table()
{
  olsr_u8_t              index;
  struct duplicatehash   *dup_hash;
  struct duplicate_entry *dup_message;
  struct duplicate_entry *dup_message_tmp;
  
  for(index=0;index<HASHSIZE;index++)
    {
      dup_hash=&duplicatetable[index];
      dup_message=dup_hash->duplicate_forw;
      while(dup_message!=(struct duplicate_entry *)dup_hash)
	{
	  if(olsr_timed_out(&dup_message->duplicate_timer))
	    {
	      dup_message_tmp=dup_message;
	      dup_message=dup_message->duplicate_forw;
	      olsr_delete_duplicate_table(dup_message_tmp);
	    }
	  else
	    dup_message=dup_message->duplicate_forw;	    
	}
    }
}
/*------------------------------------------------------------------------*/

int
olsr_print_duplicate_table()
{
  olsr_u8_t              index;
  struct duplicatehash   *dup_hash;
  struct duplicate_entry *dup_message;
  
  printf("DUPLICATE TABLE\n");
  printf("   HASH           ADDRESS          DUPLICATE seq nu      \n");
  printf("---------------------------------------------------------\n");
  
  for(index=0;index<HASHSIZE;index++)
    {
      dup_hash=&duplicatetable[index];
      
      for(dup_message=dup_hash->duplicate_forw;dup_message!=(struct duplicate_entry *)dup_hash;dup_message=dup_message->duplicate_forw)
	{
	  printf("%d              ",dup_message->duplicate_hash);	  
	  printf("%s            ",convert_address_to_string(&dup_message->duplicate_addr));
	  printf("%d            \n",dup_message->duplicate_seq);
	  
	}
    }
  printf("######################################################################\n");
}
/*------------------------------------------------------------------------*/
