/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char kernel_routes_rcsid[] = 
  "$Id: kernel_routes.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


#include "defs.h"
#define NETMASK_HOST 0xffffffff
#define NETMASK_DEFAULT 0x0


struct rthash                     mirror_table[HASHSIZE];
extern struct interf_name *interf_names;
/*------------------------------------------------------------------------*/
int
olsr_find_up_route(struct rt_entry *dst,struct rthash *table)
{ 
  struct rt_entry *destination;
  struct rthash   *routing_hash;
  olsr_u32_t      hash;
 
  olsr_hashing(&dst->rt_dst,&hash);
  routing_hash=&table[hash & HASHMASK];

  for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
    {
      if(destination->rt_hash!=hash)
	continue;
      
      if (0==memcmp(&destination->rt_dst,&dst->rt_dst,sizeof(olsr_ip_addr)))
	if (0==memcmp(&destination->rt_router,&dst->rt_router,sizeof(olsr_ip_addr)))
	  return 1;
	else
	  return 0;
    }
  return 0;

}
/*------------------------------------------------------------------------*/
struct destination_n 
*olsr_build_update_list(struct rthash *from_table,struct rthash *in_table)
{
  struct destination_n *kernel_route_list=NULL;
  struct destination_n *route_list=NULL;
  struct rt_entry      *destination;
  struct rthash        *routing_hash;
  olsr_u8_t            index;
  
  for(index=0;index<HASHSIZE;index++)
    {
      routing_hash=&from_table[index];
      for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
	{
	  
	  if (!olsr_find_up_route(destination,in_table))
	    {
	      route_list=(struct destination_n *)malloc(sizeof(struct destination_n));
	      route_list->destination=destination;
	      
	      route_list->next=kernel_route_list;
	      kernel_route_list=route_list;
	    }
	}   
    }

  return (kernel_route_list);
}
/*------------------------------------------------------------------------*/
void
olsr_delete_all_kernel_routes(int signal)
{ 
  /*
    This function will delete all entries from the kernel routing table
    One should be careful to reconfigure the default entry.
    
  */
  struct destination_n *delete_kernel_list=NULL;
  
  
  delete_kernel_list=olsr_build_update_list(routingtable,mirror_table);

  olsr_delete_routes_from_kernel(delete_kernel_list);

  PICAcleanup();

  printf("\n ******************** Olsrd: Bye Bye !!!  ********************\n");

  exit(signal);
}
/*------------------------------------------------------------------------*/
int
olsr_update_kernel_routes()
{
  struct destination_n *delete_kernel_list=NULL;
  struct destination_n *add_kernel_list=NULL;
  
  delete_kernel_list=olsr_build_update_list(mirror_table,routingtable);
  add_kernel_list=olsr_build_update_list(routingtable,mirror_table);
  olsr_delete_routes_from_kernel(delete_kernel_list);
  olsr_add_routes_in_kernel(add_kernel_list);
}
/*------------------------------------------------------------------------*/
int
olsr_routing_mirror()
{
  /*
    Cette procedure permet de copier la table de routage dans la mirror table
    avant de recalculer la table de routage elle doit etre appele a la place de l'appel
    de {olsr_release_routing_table();} qui doit etre remplacee par une autre procedure
    acceptant une tabel comme argumant
  */
  olsr_16_t            index;

  for(index=0;index<HASHSIZE;index++)
    {
      
      if(routingtable[index].rt_forw==(struct rt_entry *)&routingtable[index])
	{
	  mirror_table[index].rt_forw=(struct rt_entry *)&mirror_table[index];
	  mirror_table[index].rt_back=(struct rt_entry *)&mirror_table[index];
	}
      else
	{
	   mirror_table[index].rt_forw=routingtable[index].rt_forw;
	   mirror_table[index].rt_forw->rt_back=(struct rt_entry *)&mirror_table[index];
	   mirror_table[index].rt_back=routingtable[index].rt_back;
	   mirror_table[index].rt_back->rt_forw=(struct rt_entry *)&mirror_table[index];
	   
	   routingtable[index].rt_forw=(struct rt_entry *)&routingtable[index];
	   routingtable[index].rt_back=(struct rt_entry *)&routingtable[index];
	}       
    }

}

/*------------------------------------------------------------------------*/
int 
olsr_delete_routes_from_kernel(struct destination_n *delete_kernel_list)
{
  struct destination_n *destination_kernel;
  olsr_16_t error;

  while(delete_kernel_list!=NULL)
    {
      error=olsr_ioctl_del_route(delete_kernel_list->destination);

      /*----------------------*/
      if((error<0)&&(debug))
	printf("%s\n",strerror(errno));
      /*----------------------*/

      destination_kernel=delete_kernel_list;
      delete_kernel_list=delete_kernel_list->next;
      
      free(destination_kernel);
    }
}
/*------------------------------------------------------------------------*/
int 
olsr_add_routes_in_kernel(struct destination_n *add_kernel_list)
{
  struct destination_n *destination_kernel;
  olsr_16_t error;

  while(add_kernel_list!=NULL)
    {
      error=olsr_ioctl_add_route(add_kernel_list->destination);

      /*---------------------*/
      if((error<0)&&(debug))
	printf("%s\n",strerror(errno));
      /*---------------------*/

      destination_kernel=add_kernel_list;
      add_kernel_list=add_kernel_list->next;
      
      free(destination_kernel);
    }
}
/*------------------------------------------------------------------------*/
#undef rt_flags
#undef rt_ifp
#undef rt_metric
#undef rt_dst
/*------------------------------------------------------------------------*/
int
olsr_ioctl_add_route(struct rt_entry *destination)
{
//  struct rtentry kernel_route;
    UINT32 dest=0, mask=0, gw=0;
    int metric=0;	
#ifdef PICA_LINUX
    char device[5] = "";
#else
    char device[50] = "";
#endif

/*
  memset(&kernel_route,0,sizeof(struct rtentry));

  ((struct sockaddr_in *)&kernel_route.rt_dst)->sin_addr.s_addr=destination->rt_entry_infos.rtu_dst;
  ((struct sockaddr_in *)&kernel_route.rt_dst)->sin_family=AF_INET;
*/

    dest = destination->rt_entry_infos.rtu_dst;
  
    if(destination->rt_entry_infos.rtu_dst!=destination->rt_entry_infos.rtu_router)
    {
/*      ((struct sockaddr_in *)&kernel_route.rt_gateway)->sin_addr.s_addr=destination->rt_entry_infos.rtu_router;
	((struct sockaddr_in *)&kernel_route.rt_gateway)->sin_family=AF_INET;*/
	
	gw = destination->rt_entry_infos.rtu_router;	
    }
#ifndef PICA_LINUX
	else {
	gw = addr_ip;
	}
#endif


/*  ((struct sockaddr_in *)&kernel_route.rt_gateway)->sin_family=AF_INET;

  kernel_route.rt_flags = destination->rt_entry_infos.rtu_flags;

  ((struct sockaddr_in *)&kernel_route.rt_genmask)->sin_family=AF_INET;  */

  if(destination->rt_entry_infos.rtu_dst==INADDR_ANY)
//    ((struct sockaddr_in *)&kernel_route.rt_genmask)->sin_addr.s_addr=NETMASK_DEFAULT;
      mask = NETMASK_DEFAULT;
  else
//    ((struct sockaddr_in *)&kernel_route.rt_genmask)->sin_addr.s_addr=NETMASK_HOST;
      mask = NETMASK_HOST;
  
//  kernel_route.rt_metric=destination->rt_entry_infos.rtu_metric + 1;

  metric = destination->rt_entry_infos.rtu_metric;

  if (ifnet)
    {
/*      kernel_route.rt_dev=(char*) malloc(strlen(ifnet->int_name));
	strcpy(kernel_route.rt_dev,ifnet->int_name);*/
	strcpy(device,ifnet->int_name);
    }	

	
	if (!PICAaddRoute(dest,mask,gw,metric,device)) {
		//printf("\n\nError adding route!!!\n\n");
		return -1;
	}

/*  kernel_route.rt_dev=*/
  
  return 0;

//  return(ioctl(s,SIOCADDRT,&kernel_route));
}
/*------------------------------------------------------------------------*/
int
olsr_ioctl_del_route(struct rt_entry *destination)
{

//  struct rtentry kernel_route;
    UINT32 dest=0, mask=0, gw=0;
    int metric;
#ifdef PICA_LINUX
    char device[5] = "";
#else
    char device[50] = "";
#endif

/*  memset(&kernel_route,0,sizeof(struct rtentry));

  ((struct sockaddr_in *)&kernel_route.rt_dst)->sin_addr.s_addr=destination->rt_entry_infos.rtu_dst;
  ((struct sockaddr_in *)&kernel_route.rt_dst)->sin_family=AF_INET;

  ((struct sockaddr_in *)&kernel_route.rt_genmask)->sin_family=AF_INET;
   if(destination->rt_entry_infos.rtu_dst==INADDR_ANY)
    ((struct sockaddr_in *)&kernel_route.rt_genmask)->sin_addr.s_addr=NETMASK_DEFAULT;
  else
    ((struct sockaddr_in *)&kernel_route.rt_genmask)->sin_addr.s_addr=NETMASK_HOST;
  
   return(ioctl(s,SIOCDELRT,&kernel_route));    
*/
    
    dest = destination->rt_entry_infos.rtu_dst;
    if(destination->rt_entry_infos.rtu_dst==INADDR_ANY) mask=NETMASK_DEFAULT;
  else mask=NETMASK_HOST;
    
    if(destination->rt_entry_infos.rtu_dst!=destination->rt_entry_infos.rtu_router) 
	gw = destination->rt_entry_infos.rtu_router;
    
    if (ifnet) {
	strcpy(device,ifnet->int_name);
    }
    
	
    if (!PICAdelRoute(dest, mask,gw, device)) return -1;
    
    return 0;
}
/*------------------------------------------------------------------------*/
int
olsr_release_kernel_routing_table(struct rthash *table)
{
  struct rt_entry *destination;
  struct rthash   *routing_hash;
  olsr_u8_t             index;
  
  
  for(index=0;index<HASHSIZE;index++)
    {
      routing_hash=&table[index];
      for(destination=routing_hash->rt_forw;destination!=(struct rt_entry *)routing_hash;destination=destination->rt_forw)
        {
          olsr_ioctl_del_route(destination);
        }
    } 
  
  
}
/*-----------------------------------------------------------------------*/
