/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char mpr_selector_table_rcsid[] = 
  "$Id: mpr_selector_table.c,v 1.2 2000/12/06 10:36:11 prima Exp $";


#include "defs.h"

struct mpr_selector_table         mprstable;

/*------------------------------------------------------------------------*/
int
olsr_delete_mpr_selector_table(struct mpr_selector_entry *mprs)
{

  olsr_remque((struct olsr_qelem *)mprs);
  free((void *)mprs);
  
}
/*------------------------------------------------------------------------*/
int
olsr_insert_mpr_selector_table(struct mpr_selector_entry *mprs)
{

  olsr_u32_t  hash;
  struct mpr_selector_hash  *mprs_hash;
  
  olsr_hashing(&mprs->mpr_selector_addr,&hash);
  mprs->mprselector_hash=hash;
  mprs_hash=&mprstable.mpr_selector_hash[hash & HASHMASK];
  olsr_insque((struct olsr_qelem *)mprs, (struct olsr_qelem *)mprs_hash);
  
}

/*------------------------------------------------------------------------*/
struct mpr_selector_entry
*olsr_lookup_mpr_selector_table(olsr_ip_addr *dst)
{
  struct mpr_selector_entry *mprs;
  struct mpr_selector_hash  *mprs_hash;
  olsr_u32_t                hash;
  
   olsr_hashing(dst,&hash);
   mprs_hash=&mprstable.mpr_selector_hash[hash & HASHMASK];
  
  for(mprs=mprs_hash->mpr_selector_forw;mprs!=(struct mpr_selector_entry *)mprs_hash;mprs=mprs->mpr_selector_forw)
    {
      if(mprs->mprselector_hash!=hash)
	continue;

      if (0==memcmp(&mprs->mpr_selector_addr,dst,sizeof(olsr_ip_addr)))
	return(mprs);
    }
  return(NULL);
}
/*------------------------------------------------------------------------*/
int
olsr_update_mpr_selector_table(struct hello_message *message)
{
  /*
    If the mpr set is changed, then we must increment the mssn
  */

  struct mpr_selector_entry  *mprs;
    
  if(NULL==(mprs=olsr_lookup_mpr_selector_table(&message->source_addr)))
    {

      mprs=(struct mpr_selector_entry *)malloc(sizeof(struct mpr_selector_entry));
      memcpy(&mprs->mpr_selector_addr,&message->source_addr,sizeof(olsr_ip_addr));
      olsr_insert_mpr_selector_table(mprs);
      mprs->mpr_selector_seq=message->mpr_seq_number;
      timeradd(&now,&hold_time_neighbor,&mprs->mpr_selector_timer);
      /*increment the mssn number don't forget to increment this number on time out too*/
      mprstable.mssn++;
      /* Set the changes flag to UP ( This flag will be set to DOWN after sending a TC packet)*/
      changes=UP;
    }
  else
    {
      if(message->mpr_seq_number>=mprs->mpr_selector_seq)
	{	
	  mprs->mpr_selector_seq=message->mpr_seq_number;	  
	  timeradd(&now,&hold_time_neighbor,&mprs->mpr_selector_timer);
	}
    }
}
/*------------------------------------------------------------------------*/

int
olsr_release_mpr_seletector_table()
{
  struct mpr_selector_entry *mprs;
  struct mpr_selector_entry *mprs_tmp;
  struct mpr_selector_hash  *mprs_hash;
  olsr_u8_t                   index;

  for(index=0;index<HASHSIZE;index++)
    {
      mprs_hash=&mprstable.mpr_selector_hash[index];

      mprs=mprs_hash->mpr_selector_forw;
      while(mprs!=(struct mpr_selector_entry *)mprs_hash)
	{
	  mprs_tmp=mprs;
	  mprs=mprs->mpr_selector_forw;
	  olsr_delete_mpr_selector_table(mprs_tmp);
	}
    }
  
}
/*------------------------------------------------------------------------*/
int
olsr_time_out_mpr_selector_table()
{
  struct mpr_selector_entry *mprs;
  struct mpr_selector_entry *mprs_tmp;
  struct mpr_selector_hash  *mprs_hash;
  olsr_u8_t                 index;
  
  for(index=0;index<HASHSIZE;index++)
    {
      mprs_hash=&mprstable.mpr_selector_hash[index];
      mprs=mprs_hash->mpr_selector_forw;

      while(mprs!=(struct mpr_selector_entry *)mprs_hash)	
	{
	  if(olsr_timed_out(&mprs->mpr_selector_timer))
	    {
	      mprs_tmp=mprs;
	      mprs=mprs->mpr_selector_forw;
	      olsr_delete_mpr_selector_table(mprs_tmp);
	      
	      /*increment the mssn number don't forget to increment this number */
	      mprstable.mssn++;
	      /* The changes flag is set to UP*/
	      changes=UP;
	    }
	  else
	    mprs=mprs->mpr_selector_forw;
	}
    }
}
/*------------------------------------------------------------------------*/
int
olsr_print_mpr_selector_table()
{
  struct mpr_selector_entry *mprs;
  struct mpr_selector_hash  *mprs_hash;
  olsr_u8_t                 index;

  printf("MPR SELECTOR TABLE\n");
  
  for(index=0;index<HASHSIZE;index++)
    {
      mprs_hash=&mprstable.mpr_selector_hash[index];
      
      for(mprs=mprs_hash->mpr_selector_forw;mprs!=(struct mpr_selector_entry *)mprs_hash;mprs=mprs->mpr_selector_forw)
	{	 	  
	  printf("%s            \n",convert_address_to_string(&mprs->mpr_selector_addr));	  
	}
    }
}
/*------------------------------------------------------------------------*/
