/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Routing table management daemon.
 */

char neighbor_table_rcsid[] = 
  "$Id: neighbor_table.c,v 1.2 2000/12/06 10:36:11 prima Exp $";

#include "defs.h"


struct neighbor_table             neighbortable;

int 
olsr_delete_neighbor_2_pointer(struct neighbor_entry *neighbor,olsr_ip_addr *address)
{

  struct neighbor_2_list_entry *neighbor_two_list;
  struct neighbor_2_list_entry *neighbor_two_list_tmp;
  
  neighbor_two_list=neighbor->neighbor_2_list;

  neighbor->neighbor_2_list=NULL;

  while(neighbor_two_list!=NULL)
    {
      
      if(0==memcmp(&neighbor_two_list->neighbor_2->neighbor_2_addr,address,sizeof(olsr_ip_addr)))
	{
	  neighbor_two_list_tmp= neighbor_two_list;
	  neighbor_two_list=neighbor_two_list->neighbor_2_next;

	  free(neighbor_two_list_tmp);	  
	}
      else
	{
	  neighbor_two_list_tmp= neighbor_two_list;
	  neighbor_two_list=neighbor_two_list->neighbor_2_next;
	  neighbor_two_list_tmp->neighbor_2_next= neighbor->neighbor_2_list;
	  neighbor->neighbor_2_list=neighbor_two_list_tmp;
	}      
    }
}

/*------------------------------------------------------------------------*/
struct neighbor_2_list_entry
*olsr_lookup_my_neighbors(struct neighbor_entry *neighbor,olsr_ip_addr *neighbor_address)
{
  
  struct neighbor_2_list_entry *neighbor_two_list;
  
  neighbor_two_list=neighbor->neighbor_2_list;
  
  while(neighbor_two_list!=NULL)
    {
      
      if(0==memcmp(&neighbor_two_list->neighbor_2->neighbor_2_addr,neighbor_address,sizeof(olsr_ip_addr)))
	return(neighbor_two_list);
      
      neighbor_two_list=neighbor_two_list->neighbor_2_next;
      
    }
  return(NULL);
}
/*------------------------------------------------------------------------*/
int
olsr_delete_neighbor_table(struct neighbor_entry *neighbor)
{
  /*
    Delete a neighbor entry results the deletion of its 2 hop neighbors list!!!
   */
  
  struct  neighbor_2_list_entry *two_hop_list;
  struct  neighbor_2_entry      *two_hop_entry;

  two_hop_list = neighbor->neighbor_2_list;

  while(two_hop_list!=NULL)

    {
      
      two_hop_entry=two_hop_list->neighbor_2;
      
      two_hop_entry->neighbor_2_pointer--;

      olsr_delete_neighbor_pointer(two_hop_entry,&neighbor->neighbor_addr);
      
      if(two_hop_entry->neighbor_2_pointer<1)
	/* means this two hop entry is no more pointed, not reachable */
	{
	  olsr_remque((struct olsr_qelem *)two_hop_entry);

	  free((void *)two_hop_entry);
	}


      neighbor->neighbor_2_list=two_hop_list->neighbor_2_next;     
      free(two_hop_list);
      two_hop_list = neighbor->neighbor_2_list;
      
    }

  if(neighbor->neighbor_ifp!=NULL)    
    free(neighbor->neighbor_ifp);

  TRACE_ACTION("DELETE NEIGHBOR", &neighbor->neighbor_addr,NULL,0);
  olsr_remque((struct olsr_qelem *)neighbor);
 
  free((void *)neighbor);
}
/*------------------------------------------------------------------------*/
int
olsr_insert_neighbor_table(struct neighbor_entry *neighbor)
{

  olsr_u32_t             hash;
  struct neighborhash    *hash_neighbor;

  olsr_hashing(&neighbor->neighbor_addr,&hash);
  neighbor->neighbor_hash=hash;

  hash_neighbor=&neighbortable.neighborhash[hash & HASHMASK];

  TRACE_ACTION("ADD NEIGHBOR", &neighbor->neighbor_addr,NULL,0);
  olsr_insque((struct olsr_qelem *)neighbor, (struct olsr_qelem *)hash_neighbor);

}

/*------------------------------------------------------------------------*/
struct neighbor_entry 
*olsr_lookup_neighbor_table(olsr_ip_addr *dst)
{
  struct neighbor_entry  *neighbor;
  struct neighborhash    *hash_neighbor;
  olsr_u32_t             hash;

  olsr_hashing(dst,&hash);
  hash_neighbor=&neighbortable.neighborhash[hash & HASHMASK];

  for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
    {
      if(neighbor->neighbor_hash!=hash)
	continue;

      if (0==memcmp(&neighbor->neighbor_addr,dst,sizeof(olsr_ip_addr)))
	return(neighbor);
      
    }
  return (NULL);

  
}

/*------------------------------------------------------------------------*/

int
olsr_release_neighbor_table()
{
  olsr_u8_t              index;
  struct neighbor_entry  *neighbor;
  struct neighborhash    *hash_neighbor;

  for(index=0;index<HASHSIZE;index++)
    {
      hash_neighbor=&neighbortable.neighborhash[index];

      for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
	{
	  olsr_delete_neighbor_table(neighbor);
	}
    }
}
/*------------------------------------------------------------------------*/

int
olsr_time_out_two_hop_neighbors(struct neighbor_entry  *neighbor)
{
  struct  neighbor_2_list_entry *two_hop_list;
  struct  neighbor_2_list_entry *two_hop_list_tmp;
  struct  neighbor_2_entry      *two_hop_entry;


  two_hop_list = neighbor->neighbor_2_list;
  neighbor->neighbor_2_list=NULL;
  while(two_hop_list!=NULL)
    {
      if(olsr_timed_out(&two_hop_list->neighbor_2_timer))
	{
	  
	  two_hop_entry=two_hop_list->neighbor_2;
	  two_hop_entry->neighbor_2_pointer--;
	  olsr_delete_neighbor_pointer(two_hop_entry,&neighbor->neighbor_addr);
      
	  if(two_hop_entry->neighbor_2_pointer<1)
	    /* means this two hop entry is no more pointed, not reachable */
	    {	    
	      olsr_remque((struct olsr_qelem *)two_hop_entry);	      
	      free((void *)two_hop_entry);
	    }
	  two_hop_list_tmp=two_hop_list;
	  two_hop_list=two_hop_list->neighbor_2_next;	  
	  free(two_hop_list_tmp);

	  /* This flag is set to UP to recalculate the MPR set and the routing table*/
	  changes_neighborhood=UP;
	  changes_topology=UP;
	  
	}
      else
	{
	  two_hop_list_tmp=two_hop_list;
	  two_hop_list=two_hop_list->neighbor_2_next;
	  two_hop_list_tmp->neighbor_2_next=neighbor->neighbor_2_list;
	  neighbor->neighbor_2_list=two_hop_list_tmp;
	}
    }
}
/*------------------------------------------------------------------------*/
int
olsr_time_out_neighborhood_tables()
{
  olsr_u8_t              index;
  struct neighbor_entry  *neighbor;
  struct neighbor_entry  *neighbor_tmp;
  struct neighborhash    *hash_neighbor;
  
  for(index=0;index<HASHSIZE;index++)
    {
      hash_neighbor=&neighbortable.neighborhash[index];
      neighbor=hash_neighbor->neighbor_forw;

      while(neighbor!=(struct neighbor_entry *)hash_neighbor)
	{ 

	  /*---------------*/
	  if(debug)
	    {
	      printf("timeout: %d   %d\n ",neighbor->neighbor_timer.tv_sec,neighbor->neighbor_timer.tv_usec);
	      printf("now: %d   %d\n ",now.tv_sec,now.tv_usec);
	    }
	  /*--------------*/

	  if(olsr_timed_out(&neighbor->neighbor_timer))
	    {
	      if((neighbor->neighbor_status==NBS_ASYM) ||(neighbor->neighbor_status==NBS_LOST))
		{
		  /*-----------------*/
		  if(debug)
		    printf("deleting neighbor entry\n");
		  /*-----------------*/

		  neighbor_tmp=neighbor;
		  neighbor=neighbor->neighbor_forw;

		  olsr_delete_neighbor_table(neighbor_tmp);
		  /* This flag is set to UP to recalculate the MPR set and the routing table*/
		  changes_neighborhood=UP;
		  changes_topology=UP;
		}
	      else
		{
		  neighbor->neighbor_status=NBS_LOST;
		  neighbor=neighbor->neighbor_forw;
		}
	    }
	  else
	    {
	      olsr_time_out_two_hop_neighbors(neighbor);
	      neighbor=neighbor->neighbor_forw;
	    }
	}
    }
}
/*------------------------------------------------------------------------*/
int
olsr_print_neighbor_table()
{
  olsr_u8_t                    index;
  struct neighbor_entry        *neighbor;
  struct neighborhash          *hash_neighbor;
  struct neighbor_2_list_entry *list_2;

  for(index=0;index<HASHSIZE;index++)
    {
      hash_neighbor=&neighbortable.neighborhash[index];
      
      for(neighbor=hash_neighbor->neighbor_forw;neighbor!=(struct neighbor_entry *)hash_neighbor;neighbor=neighbor->neighbor_forw)
	{
	  
	  printf("ADDRESS        STATUS\n"); 
	  printf("------------------------- \n");
	  printf("%s                 ",convert_address_to_string(&neighbor->neighbor_addr));
	  printf("%d                 \n",neighbor->neighbor_status);
	  printf("TWO HOP NEIGHBORS ADDRESS\n");
	  printf("------------------------- \n");
	  list_2=neighbor->neighbor_2_list;
	  while(list_2!=NULL)
	    {
	      
	      printf("%s \n",convert_address_to_string(&list_2->neighbor_2->neighbor_2_addr));
	      list_2=list_2->neighbor_2_next;
	    }
	  printf("------------------------- \n");
	}
    }
}
/*------------------------------------------------------------------------*/
