/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * From: @(#)output.c	5.15 (Berkeley) 2/28/91
 * From: @(#)output.c	8.1 (Berkeley) 6/5/93
 * From: @(#)output.c	1.6 (dholland Exp)  1996/11/25
 */

char output_rcsid[] = 
  "$Id: output.c,v 1.2 2000/12/06 10:36:12 prima Exp $";


/*
 * Routing Table Management Daemon
 */


#include "defs.h"


/*
 * Apply the function "f" to all non-passive
 * interfaces.Use broadcast address.
 */

void toall(void (*f)(struct sockaddr *, struct hello_message *, 
                     struct tc_message *, int, struct _interface *), 
           struct hello_message *hmsg, struct tc_message *tmsg,
           struct _interface *skipif)
{
	struct _interface *ifp;
	struct sockaddr *dst;
	int flags;

	for (ifp = ifnet; ifp; ifp = ifp->int_next) {
		if (ifp->int_flags & IFF_PASSIVE || ifp == skipif)
			continue;
		dst = ifp->int_flags & IFF_BROADCAST ? &ifp->int_broadaddr :
		      &ifp->int_addr;
		flags = ifp->int_flags & IFF_INTERFACE ? MSG_DONTROUTE : 0;

		(*f)(dst, hmsg, tmsg, flags, ifp);
	}
}

/*
 * Output a preformed packet.
 * The two parameters message and messag are not used.
 */

void sndmsg(struct sockaddr *dst, struct hello_message *message, 
            struct tc_message *messag, int flags, struct _interface *ifp)
{

	TRACE_OUTPUT(ifp, dst, outputsize);
	(*afswitch[dst->sa_family].af_output)(s, flags,
		dst, outputsize);
}


/*
 * Supply dst with the contents of the routing tables or neighbor table.
 * If this won't fit in one packet, chop it up into several.
 */
 
void supply(struct sockaddr *dst, struct hello_message *h_packet,
                    struct rt_message *r_packet, int type)
{
    struct _interface *ifp = 0;

        if ( outputsize )
	       toall(sndmsg, (struct hello_message *)NULL, 
                     (struct tc_message *)NULL, (struct _interface *)NULL );

	if (type == HELLO_REQUEST)
	  hello_build(dst, h_packet, (struct tc_message *)NULL, 0, 
                        (struct _interface *)NULL);
        else{
    	   if (type == ROUTE_REQUEST)
	     rt_build(dst, r_packet, 0, (struct _interface *)NULL);
	}
	if ( outputsize ){
           TRACE_OUTPUT(ifp, dst, outputsize);
	   (*afswitch[dst->sa_family].af_output)(s, 0, dst, outputsize);
	}

}


/*
 * Generate HELLO packet with the contents of the parameter "message".
 * If this won't fit in one packet, chop it up into several.
 * Send the packet if the size of the data contained in the output buffer
 * reach MAXPACKETSIZE. Can generate an empty HELLO packet if the 
 * neighbor table is empty. The parameter messag is not used.
 */
 
void hello_build(struct sockaddr *dst, struct hello_message *message, 
                 struct tc_message *messag, int flags, struct _interface *ifp)
{

     void (*output)(int,int,struct sockaddr *,int) = 
                                          afswitch[dst->sa_family].af_output;

     int remainsize = outputsize - (sizeof(msg->olsr_packlen) 
                                    + sizeof(msg->olsr_res));
     struct hello_neighbor *nb, *prev_nb;

     struct olsrmsg *m = outputsize ? 
                      (struct olsrmsg *)((char *)msg->olsr_msg + remainsize)
                      : msg->olsr_msg;
     struct hellomsg *h = m->olsr_hello;
     struct hellinfo *hinfo = h->hell_info;
     olsr_ip_addr *haddr = hinfo->neigh_addr;

     int  i, sametype, npackets = 0, found =0;
     int lastpacket = 0; /* number of neighbors with the same
        		   greater link status in the last packet */

     if (!message)
	  return;
     
	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) {
	      msg->olsr_packlen = htons(outputsize);
	      memset(msg->olsr_res, 0, sizeof(msg->olsr_res));

	      TRACE_OUTPUT(ifp, dst, outputsize);
	      (*output)(s, flags, dst, outputsize);
	}

        for (i = 1; i <= HELLOSTATUS_MAX; i++) {

	    lastpacket = sametype = 0;
            outputsize = (char *)hinfo - packet;
	    if (outputsize > MAXPACKETSIZE - (int)sizeof (struct hellinfo)) {

                    msg->olsr_packlen = htons(outputsize);
	            memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
                    m->olsr_msgtype = HELLO_PACKET;
                    m->olsr_broad = MSG_NOBROADCAST;
                    m->olsr_nextmsg = (char *)hinfo - (char *)m;
                    m->olsr_nextmsg = htons(m->olsr_nextmsg);
                    hinfo->next_linktype = (char *)haddr - (char *)hinfo;
                    hinfo->next_linktype = ntohs(hinfo->next_linktype);
                    h->hell_seqnum = htons(hello_seqnum++);  
                    h->hell_mprseqnum = htons(message->mpr_seq_number);

	            TRACE_OUTPUT(ifp, dst, outputsize);
		    (*output)(s, flags, dst, outputsize);

                    m = msg->olsr_msg;
                    h = m->olsr_hello;
                    hinfo = h->hell_info;
                    haddr = hinfo->neigh_addr;
                    npackets++;	
	    }          
 	  
	    for (nb = message->neighbors; nb != NULL;
                 nb = nb->next) {

		outputsize = (char *)haddr - packet;
		if (outputsize > MAXPACKETSIZE - (int)sizeof (haddr))
                   {

	                msg->olsr_packlen = htons(outputsize);
	                memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
                        m->olsr_msgtype = HELLO_PACKET;
                        m->olsr_broad = MSG_NOBROADCAST;
                        m->olsr_nextmsg = (char *)haddr - (char *)m;
                        m->olsr_nextmsg = htons(m->olsr_nextmsg);
                        hinfo->next_linktype = (char *)haddr - (char *)hinfo;
                        hinfo->next_linktype = ntohs(hinfo->next_linktype);
                        h->hell_seqnum = htons(hello_seqnum++);
                        h->hell_mprseqnum = htons(message->mpr_seq_number);

			TRACE_OUTPUT(ifp, dst, outputsize);
			(*output)(s, flags, dst, outputsize);

                        m = msg->olsr_msg;
                        h = m->olsr_hello;
                        hinfo = h->hell_info;
                        haddr = hinfo->neigh_addr;
                        hinfo->link_type = i;
                        lastpacket = sametype;
                        npackets++;
		}
                if (nb->status == i){
                   sametype++;
                   if (sametype == 1){
		     hinfo->link_type = i;
	             memset(hinfo->res, 0, sizeof(hinfo->res));
                   }
           	   *haddr = nb->address;
                   found = 1;

	           haddr++;
                }
	    }
       
        if (sametype && sametype > lastpacket){
           hinfo->next_linktype = (char *)haddr - (char *)hinfo;
           hinfo->next_linktype = ntohs(hinfo->next_linktype);
           hinfo = (struct hellinfo *)((char *)haddr);
           haddr = hinfo->neigh_addr;
        }

        } /* for i*/

        h->hell_seqnum = htons(hello_seqnum++);
        h->hell_mprseqnum = htons(message->mpr_seq_number);
        m->olsr_msgtype = HELLO_PACKET;
        m->olsr_broad = MSG_NOBROADCAST;
        m->olsr_nextmsg = (char *)hinfo - (char *)m;
        m->olsr_nextmsg = htons(m->olsr_nextmsg);
	outputsize = (char *)hinfo - packet;
        msg->olsr_packlen = htons(outputsize);
	memset(msg->olsr_res, 0, sizeof(msg->olsr_res));

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) {

           TRACE_OUTPUT(ifp, dst, outputsize);
	   (*output)(s, flags, dst, outputsize);

	}

       /*
        * Delete the list of neighbor messages.
        */

        nb = message->neighbors;

	while (nb)
	  {
	    prev_nb = nb;
	    nb = nb->next;
	    free(prev_nb);
	  }

}


/*
 * Generate TC packet with the contents of the parameter "message".
 * If this won't fit in one packet, chop it up into several.
 * Send the packet if the size of the data contained in the output buffer
 * reach MAXPACKETSIZE. Don't generate an empty TC packet.
 * The parameter messag is not used.
 */
 
void tc_build(struct sockaddr *dst, struct hello_message *messag,
              struct tc_message *message, int flags, struct _interface *ifp)           
{

     void (*output)(int,int,struct sockaddr *,int) = 
                                          afswitch[dst->sa_family].af_output;

     int remainsize = outputsize - (sizeof(msg->olsr_packlen) 
                                 + sizeof(msg->olsr_res));
     struct tc_mpr_addr *mprs, *prev_mprs;
     struct olsrmsg *m = outputsize ? 
                       (struct olsrmsg *)((char *)msg->olsr_msg + remainsize)
                       : msg->olsr_msg;
     struct tcmsg *tc = m->olsr_tc;
     olsr_ip_addr *mprsaddr = tc->tc_mprsaddr; 
     struct _interface *interf;
     struct sockaddr address;
     int found = 0;

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) {
	      msg->olsr_packlen = htons(outputsize);
              memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
 
	      TRACE_OUTPUT(ifp, dst, outputsize);
	      (*output)(s, flags, dst, outputsize);
	}
            
        if (!message)
	    return;

        memset(address.sa_data, 0, sizeof(address.sa_data));
        ((struct sockaddr_in *)&address)->sin_addr.s_addr = 
	  message->originator;
        address.sa_family = AF_INET;
        interf = if_ifwithaddr(&address);

	for (mprs = message->multipoint_relay_selector_address; 
             mprs != NULL;
             mprs = mprs->next) {

	       outputsize = (char *)mprsaddr - packet;
	       if (outputsize > MAXPACKETSIZE - (int)sizeof (mprsaddr)){

	          msg->olsr_packlen = htons(outputsize);
                  memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
                  m->olsr_msgtype = TC_PACKET;
                  m->olsr_broad = MSG_BROADCAST;
                  m->olsr_nextmsg = (char *)mprsaddr - (char *)m;
                  m->olsr_nextmsg = htons(m->olsr_nextmsg);
		  if (interf)            
                     tc->tc_seqnum = htons(tc_seqnum++);
                  else
                     tc->tc_seqnum = htons(message->packet_seq_number);
                  tc->tc_mprseqnum = htons(message->mssn);
                  tc->tc_hopcnt = message->hop_count;
	          memset(tc->tc_res, 0, sizeof(tc->tc_res));
                  tc->tc_origaddr = message->originator;

		  TRACE_OUTPUT(ifp, dst, outputsize);

		  (*output)(s, flags, dst, outputsize);

                  m = msg->olsr_msg;
                  tc = m->olsr_tc;
                  mprsaddr = tc->tc_mprsaddr;
	          found = 0;
		}
	       found = 1;
 
           	*mprsaddr = mprs->address;

		mprsaddr++;
        }

        if (found){

	outputsize = (char *)mprsaddr - packet;
	msg->olsr_packlen = htons(outputsize);
        memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
        m->olsr_msgtype = TC_PACKET;
        m->olsr_broad = MSG_BROADCAST;
        m->olsr_nextmsg = (char *)mprsaddr - (char *)m;
        m->olsr_nextmsg = htons(m->olsr_nextmsg);
	if (interf)            
            tc->tc_seqnum = htons(tc_seqnum++);
        else
            tc->tc_seqnum = htons(message->packet_seq_number);
        tc->tc_mprseqnum = htons(message->mssn);
        tc->tc_hopcnt = message->hop_count;
	memset(tc->tc_res, 0, sizeof(tc->tc_res));
        tc->tc_origaddr = message->originator;

        }

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) {

           TRACE_OUTPUT(ifp, dst, outputsize);
	   (*output)(s, flags, dst, outputsize);
	}

/*
 * Delete the list of mprs messages
 */

	mprs = message->multipoint_relay_selector_address;

	while (mprs)
	  {
	    prev_mprs = mprs;
	    mprs = mprs->next;
	    free(prev_mprs);
	  }


}


/*
 * Generate a response packet to ROUTE REQUEST with the contents 
 * of the parameter "message". If this won't fit in one packet,
 * chop it up into several. Send the packet if the size of data 
 * contained in the output buffer reach MAXPACKETSIZE. 
 */
 
void rt_build(struct sockaddr *dst, struct rt_message *message, int flags,
	      struct _interface *ifp)           
{


     void (*output)(int,int,struct sockaddr *,int) = 
                                          afswitch[dst->sa_family].af_output;

     int remainsize = outputsize - (sizeof(msg->olsr_packlen) 
                                 + sizeof(msg->olsr_res));
     struct olsrmsg *m = outputsize ? 
                       (struct olsrmsg *)((char *)msg->olsr_msg + remainsize)
                       : msg->olsr_msg;
     struct rtmsg *rt = m->olsr_rt;
     struct rt_message *hostrt, *prev_hostrt;

        if (!message)
	    return;

	for (hostrt = message; hostrt != NULL; hostrt = hostrt->next) {

	       outputsize = (char *)rt - packet;
	       if (outputsize > MAXPACKETSIZE - (int)sizeof (struct rtmsg)) {

	          msg->olsr_packlen = htons(outputsize);
                  memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
                  m->olsr_msgtype = ROUTE_RESPONSE;
                  m->olsr_broad = MSG_NOBROADCAST;
                  m->olsr_nextmsg = (char *)rt - (char *)m;
                  m->olsr_nextmsg = htons(m->olsr_nextmsg);

		  TRACE_OUTPUT(ifp, dst, outputsize);

		  (*output)(s, flags, dst, outputsize);

                  m = msg->olsr_msg;
                  rt = m->olsr_rt;
		}

	       rt->rt_hopcnt = hostrt->hop_count;
               memset(rt->rt_res, 0, sizeof(rt->rt_res));

	       rt->rt_destaddr = hostrt->destination;

	       rt->rt_nextaddr = hostrt->next_hop;

	       rt++;
	}

	outputsize = (char *)rt - packet;
	msg->olsr_packlen = htons(outputsize);
        memset(msg->olsr_res, 0, sizeof(msg->olsr_res));
        m->olsr_msgtype = ROUTE_RESPONSE;
        m->olsr_broad = MSG_NOBROADCAST;
        m->olsr_nextmsg = (char *)rt - (char *)m;
        m->olsr_nextmsg = htons(m->olsr_nextmsg);

	if (outputsize > MAXPACKETSIZE - (int)sizeof (struct olsrmsg)) {

        TRACE_OUTPUT(ifp, dst, outputsize);

        (*output)(s, flags, dst, outputsize);

	}

/*
 * Delete the list of routing informations
 */

	hostrt = message;

	while (hostrt)
	  {
	    prev_hostrt = hostrt;
	    hostrt = hostrt->next;
	    free(prev_hostrt);
	  }

}
