/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 *
 * 	$Id: routing_table.h,v 1.2 2000/12/06 10:36:12 prima Exp $
 */

/*
 * Routing table management daemon.
 */


struct rthash {
	struct	rt_entry *rt_forw;
	struct	rt_entry *rt_back;
};
struct rt_entry_info{
        olsr_u32_t         rtu_hash;
        olsr_ip_addr       rtu_dst;
        olsr_ip_addr       rtu_router; 
        olsr_u8_t  	   rtu_flags;//Currently not used 
        olsr_u8_t          rtu_state;//Currently not used 
        olsr_u16_t 	   rtu_metric;
        olsr_u16_t         rtu_ifmetric;//Currently not used 
        struct	_interface *rtu_ifp;//Currently not used 
};
struct rt_entry {
	struct	rt_entry *rt_forw;
	struct	rt_entry *rt_back;
        struct  rt_entry_info rt_entry_infos;
};

#define	rt_hash		rt_entry_infos.rtu_hash	        /* for host */
#define	rt_dst		rt_entry_infos.rtu_dst	        /* Destination*/
#define	rt_router	rt_entry_infos.rtu_router	/* who to forward to */
#define	rt_flags	rt_entry_infos.rtu_flags	/* kernel flags */
#define	rt_state	rt_entry_infos.rtu_state	/* Currently not used*/
#define	rt_metric	rt_entry_infos.rtu_metric	/* cost of route */
#define	rt_ifmetric	rt_entry_infos.rtu_ifmetric	/* cost of route if, currently not used */
#define	rt_ifp		rt_entry_infos.rtu_ifp	        /* interface to take, currently not used */


/*
 * "State" of routing table entry. Currently not used
 */
#define	RTS_CHANGED	0x1		/* route has been altered recently */
#define	RTS_EXTERNAL	0x2		/* extern info, not installed or sent */
#define	RTS_INTERNAL	0x4		/* internal route, not installed */
#define	RTS_PASSIVE	IFF_PASSIVE	/* don't time out route */
#define	RTS_INTERFACE	IFF_INTERFACE	/* route is for network interface */
#define	RTS_REMOTE	IFF_REMOTE	/* route is for ``remote'' entity */
#define	RTS_SUBNET	IFF_SUBNET	/* route is for network subnet */

/*
 * Flags are same as kernel, with this addition for af_rtflags:
 */
#define	RTF_SUBNET	0x80000		/* pseudo: route to subnet */


struct destination_n{
    
    struct rt_entry      *destination;
    struct destination_n *next;
    
  };

struct rt_entry
*olsr_lookup_routing_table(olsr_ip_addr *dst);

int
olsr_delete_routing_table(struct rt_entry* destination);

int
olsr_release_routing_table(struct rthash *table);

struct rt_entry 
*olsr_insert_routing_table(olsr_ip_addr *dst);

struct rt_entry 
*olsr_insert_routing_table_from_topo(struct rt_entry *r_last,struct topology_destination_entry *destination);

struct destination_n
*olsr_fill_routing_table_with_neighbors();

int 
olsr_calculate_routing_table();

int
olsr_print_trace_routing_table(struct rthash *table);

int
olsr_print_routing_table(struct rthash *table);

int
olsr_update_default_gateway();


