/*****************************************************************************
 *
 * 2002 Polytechnic University of Valencia.
 *
 * This program was ported to the Windows 2000 / Windows CE 3.0 platforms
 * by Carlos Miguel Tavares Calafate.
 *
 * For licencing information, see the licence provided by INRIA, which is
 * a Berkeley Copyright licence.
 *
 * For more info on the Windows 2000 and Windows CE versions, check the site:
 *
 * http://reptar.grc.upv.es/~calafate/
 *
 * Authors: Carlos Calafate, <calafate@disca.upv.es>
 *
 *****************************************************************************/

/*
 * This Copyright notice is in French. An English summary is given
 * but the referee text is the French one.
 *
 * Copyright (c) 2000, 2001 Adokoe.Plakoo@inria.fr, INRIA Rocquencourt,
 *                          Anis.Laouiti@inria.fr, INRIA Rocquencourt.
 *
 * Ce logiciel informatique est disponible aux conditions
 * usuelles dans la recherche, c'est--dire qu'il peut
 * tre utilis, copi, modifi, distribu  l'unique
 * condition que ce texte soit conserv afin que
 * l'origine de ce logiciel soit reconnue.
 * Le nom de l'Institut National de Recherche en Informatique
 * et en Automatique (INRIA), ou d'une personne morale
 * ou physique ayant particip  l'laboration de ce logiciel ne peut
 * tre utilis sans son accord pralable explicite.
 * 
 * Ce logiciel est fourni tel quel sans aucune garantie,
 * support ou responsabilit d'aucune sorte.
 * Certaines parties de ce logiciel sont drives de sources developpees par
 * University of California, Berkeley et ses contributeurs couvertes 
 * par des copyrights.
 * This software is available with usual "research" terms
 * with the aim of retain credits of the software. 
 * Permission to use, copy, modify and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and the name of INRIA, or any contributor not be used in advertising
 * or publicity pertaining to this material without the prior explicit
 * permission. The software is provided "as is" without any
 * warranties, support or liabilities of any kind.
 * This product includes software developed by the University of
 * California, Berkeley and its contributors protected by copyrights.
 */

/*
 * Copyright (c) 1983, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * From: @(#)trace.c	5.11 (Berkeley) 2/28/91
 * From: @(#)trace.c	8.1 (Berkeley) 6/5/93
 */

char trace_rcsid[] = 
  "$Id: trace.c,v 1.3 2000/12/08 15:26:07 prima Exp $";


/*
 * Routing Table Management Daemon
 */
#define	OLSRTYPES
#include "defs.h"
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#ifdef PICA_LINUX
#include <syslog.h>
#include <paths.h>
#endif
#include <errno.h>


#define	NRECORDS	50		/* size of circular trace buffer */

FILE	*ftrace = NULL /*stdout*/;	/* output trace file */
int traceactions=0;			/* on/off */
int tracepackets;		/* watch packets as they go by */
int tracehistory;		/* on/off */
static int tracecontents;	/* watch packet contents as they go by */

static	struct timeval lastlog;
static	char *savetracename;
char * date_str;


static int iftraceinit(struct _interface *, struct ifdebug *);
void dumpif(FILE *, struct _interface *);
void dumptrace(FILE *, char *, struct ifdebug *);

void traceinit(struct _interface *ifp)
{

	if (iftraceinit(ifp, &ifp->int_input) &&
	    iftraceinit(ifp, &ifp->int_output))
		return;
	tracehistory = 0;
	fprintf(stderr, "traceinit: can't init %s\n", ifp->int_name);
}

static int iftraceinit(struct _interface *ifp, struct ifdebug *ifd)
{
	struct iftrace *t;

	ifd->ifd_records =
	  (struct iftrace *)malloc(NRECORDS * sizeof (struct iftrace));
	if (ifd->ifd_records == 0)
		return (0);
	ifd->ifd_front = ifd->ifd_records;
	ifd->ifd_count = 0;
	for (t = ifd->ifd_records; t < ifd->ifd_records 
	                               + NRECORDS; t++) {
		t->ift_size = 0;
		t->ift_packet = 0;
	}
	ifd->ifd_if = ifp;
	return (1);
}

void traceon(char *file)
{
    FDesc fd;
	
	if (ftrace != NULL)
		return;

	
	if (!PICAopenFile(&fd, file,WRITEF, APPEND)) {
	    sprintf(logdata, "Cannot syslog to %s: error %s\n", file, strerror(errno));
	    LOG(logdata);
	    return;
	}	
	
//	fcntl(fd, F_SETFL, 0);	/* Turn NDELAY off */
	savetracename = file;


	gettime2(&now,&timer_now);

#ifdef PICA_LINUX
	ftrace = fdopen(fd, "a");
	if (ftrace == NULL)
		return;
	dup2(fileno(ftrace), 1);
	dup2(fileno(ftrace), 2);
#endif
#ifdef PICA_WINDOWS_NT
	_dup2(fd, _fileno(stdout));
	_dup2(fd, _fileno(stderr));
#endif

	traceactions = 1;


	if (!PICAgetTimeStr(&date_str)) 
	    fprintf(stderr,"Unable to get current time string\n");	
	fprintf(ftrace, "Tracing enabled %s\n", date_str);
}

void traceoff(void)
{
#ifdef PICA_LINUX
	int fd;
#endif

	if (!traceactions)
		return;
	if (ftrace != NULL) {

	    if (!PICAgetTimeStr(&date_str)) 
		    fprintf(stderr,"Unable to get current time string\n"); 
		else {
		    fprintf(ftrace, "Tracing disabled %s\n", date_str);
                    //free(date_str);
		    fflush(ftrace);
		}
	    fclose(ftrace);
	    ftrace = NULL;

#ifdef PICA_LINUX
		fd = open(_PATH_DEVNULL, O_RDWR);
		(void) dup2(fd, 1);
		(void) dup2(fd, 2);
		(void) close(fd);
#else
		CloseHandle(GetStdHandle(STD_OUTPUT_HANDLE));
		CloseHandle(GetStdHandle(STD_ERROR_HANDLE));
#endif	
	
	}

	traceactions = 0;
	tracehistory = 0;
	tracepackets = 0;
	tracecontents = 0;
}

/*
void sigtrace(int s)
{

	if (s == SIGUSR2)
		traceoff();
	else if (ftrace == NULL && savetracename)
		traceon(savetracename);
	else
		bumploglevel();
}
*/

/*
 * Move to next higher level of tracing when -t option processed or
 * SIGUSR1 is received.  Successive levels are:
 *	traceactions
 *	traceactions + tracepackets
 *	traceactions + tracehistory (packets and contents after change)
 *	traceactions + tracepackets + tracecontents
 */

void bumploglevel(void)
{

    gettime2(&now,&timer_now);
	if (traceactions == 0) {
		traceactions++;
		if (ftrace) {
		    if (!PICAgetTimeStr(&date_str)) 
			fprintf(stderr,"Unable to get current time string\n"); 
		    fprintf(ftrace, "Tracing disabled %s\n", date_str);
		}
	} else if (tracepackets == 0) {
		tracepackets++;
		tracehistory = 0;
		tracecontents = 0;
		if (ftrace) {
		    if (!PICAgetTimeStr(&date_str)) 
			fprintf(stderr,"Unable to get current time string\n"); 
		    fprintf(ftrace, "Tracing packets started %s\n", date_str);
		}
	} else if (tracehistory == 0) {
		tracehistory++;
		if (ftrace) {
		    if (!PICAgetTimeStr(&date_str)) 
			fprintf(stderr,"Unable to get current time string\n");
		    fprintf(ftrace, "Tracing history started %s\n", date_str);
		}
	} else {
		tracepackets++;
		tracecontents++;
		tracehistory = 0;
		if (ftrace) {
		    if (!PICAgetTimeStr(&date_str)) 
			fprintf(stderr,"Unable to get current time string\n"); 
		    fprintf(ftrace, "Tracing packet contents started %s\n", date_str);
		}
	}
	if (ftrace)
		fflush(ftrace);
}

void trace(struct ifdebug *ifd, struct sockaddr *who, 
	   char *p, int len, int m)
{
	struct iftrace *t;

	if (ifd->ifd_records == 0)
		return;
	t = ifd->ifd_front++;
	if (ifd->ifd_front >= ifd->ifd_records + NRECORDS)
		ifd->ifd_front = ifd->ifd_records;
	if (ifd->ifd_count < NRECORDS)
		ifd->ifd_count++;
	if (t->ift_size > 0 && t->ift_size < len && t->ift_packet) {
		free(t->ift_packet);
		t->ift_packet = 0;
	}
	t->ift_stamp = now;
	t->ift_who = *who;
	if (len > 0 && t->ift_packet == 0) {
		t->ift_packet = (char *)malloc(len);
		if (t->ift_packet == 0)
			len = 0;
	}
	if (len > 0)
		memcpy(t->ift_packet, p, len);
	t->ift_size = len;
	t->ift_metric = m;
 
}


void traceaction(FILE *fd, char *action, olsr_ip_addr *addr,
                 struct _interface *ifp, int timer)
{

	if (fd == NULL)
		return;
	if (lastlog.tv_sec != now.tv_sec || lastlog.tv_usec != now.tv_usec) {
	    if (!PICAgetTimeStr(&date_str)) 
		fprintf(stderr,"Unable to get current time string\n");	       
	    fprintf(fd, "\n%.19s:\n", date_str);
	    lastlog = now;
	}
	fprintf(fd, "%s ", action);
	fprintf(fd, "address : %s, ", convert_address_to_string(addr));

	fprintf(fd, " timer %d\n", timer);
	if (tracehistory && !tracepackets && ifp)
		dumpif(fd, ifp);
	fflush(fd);
	if (ferror(fd))
		traceoff();
}


void dumpif(FILE *fd, struct _interface *ifp)
{
	if (ifp->int_input.ifd_count || ifp->int_output.ifd_count) {
		fprintf(fd, "*** Packet history for interface %s ***\n",
			ifp->int_name);
		dumptrace(fd, "from", &ifp->int_input);
		fprintf(fd, "*** end packet history ***\n");
	}
}

void dumptrace(FILE *fd, char *dir, struct ifdebug *ifd)
{
	struct iftrace *t;
	const char *cp = !strcmp(dir, "to") ? "Output" : "Input";

	if (ifd->ifd_front == ifd->ifd_records &&
	    ifd->ifd_front->ift_size == 0) {
		fprintf(fd, "%s: no packets.\n", cp);
		fflush(fd);
		return;
	}
	fprintf(fd, "%s trace:\n", cp);
	t = ifd->ifd_front - ifd->ifd_count;
	if (t < ifd->ifd_records)
		t += NRECORDS;
	for ( ; ifd->ifd_count; ifd->ifd_count--, t++) {
		if (t >= ifd->ifd_records + NRECORDS)
			t = ifd->ifd_records;
		if (t->ift_size == 0)
			continue;
		dumppacket(fd, dir, &t->ift_who,
		    t->ift_packet, t->ift_size, &t->ift_stamp);
	}
}

void dumppacket(FILE *fd, char *dir, struct sockaddr *sa, 
	char *cp, int size, struct timeval *stamp)
{
	struct olsr *msg = (struct olsr *)cp;
        struct olsrmsg *m = msg->olsr_msg, *mes;
        struct hellomsg *h;
	struct hellinfo *hinf;
        olsr_ip_addr *haddr, *hadr;
        struct in_addr address;
        struct tcmsg *tc;
        olsr_ip_addr *maddr;
	struct sockaddr_in *who = (struct sockaddr_in *)sa;

	if (fd == NULL)
		return;

	if (tracepackets && tracecontents == 0) {
		fflush(fd);
		return;
	}

       fprintf(fd, "\n********** Packet length:  %d **********\n",
	       ntohs(msg->olsr_packlen));

        for (mes = m; (char *)mes < (char *)msg + ntohs(msg->olsr_packlen); 
             mes = (struct olsrmsg *)((char *)mes + ntohs(mes->olsr_nextmsg))
            ) {	

       fprintf(fd, "\n             --------------------         ");
       
       if (mes->olsr_msgtype && mes->olsr_msgtype < OLSRTYPE_MAX) {
	   if (!PICAgetTimeStr(&date_str)) 
	       fprintf(stderr,"Unable to get current time string\n");	       
	   fprintf(fd, "\n %s %s %s %.19s:\n", olsrtypes[mes->olsr_msgtype], dir, inet_ntoa(who->sin_addr), date_str);
       }
	else {
	    if (!PICAgetTimeStr(&date_str)) 
		fprintf(stderr,"Unable to get current time string\n");	       
	    fprintf(fd, "Bad packet type  0x%x %s %s %.19s\n", mes->olsr_msgtype, dir, inet_ntoa(who->sin_addr), date_str);
	    
	    fprintf(fd, "size=%d cp=%p packet=%p\n", size, cp, packet);
	    fflush(fd);
	    return;
	}

          fprintf(fd, "Message type:  %d \n", mes->olsr_msgtype);
          fprintf(fd, "Broadcast ?:   %s \n", mes->olsr_broad ? "YES" : "NO");
          fprintf(fd, "Next message:  %d \n\n", ntohs(mes->olsr_nextmsg));

           switch (mes->olsr_msgtype) {

           case HELLO_PACKET:
           case UPDATE_NEIGHBOR:

	    h = mes->olsr_hello;
            fprintf(fd, "Packet seq number:  %d \n", ntohs(h->hell_seqnum));
            fprintf(fd, "MPR Sequence number:  %d \n\n", 
                         ntohs(h->hell_mprseqnum));

            for (hinf = h->hell_info; (char *)hinf < (char *)mes 
                                              + ntohs(mes->olsr_nextmsg); 
                 hinf = (struct hellinfo *)((char *)hinf 
                        + ntohs(hinf->next_linktype))
                )
               {

                fprintf(fd, "Link type:  %d \n", hinf->link_type);
                fprintf(fd, "Next link type:  %d \n", 
                             ntohs(hinf->next_linktype));

                 haddr = hinf->neigh_addr;  
                 for (hadr = haddr; (char *)hadr < (char *)hinf 
                                              + ntohs(hinf->next_linktype);
	            hadr++)
                 {
		  address.s_addr = *hadr;

		  fprintf(fd, "Neighbor addr %s\n", inet_ntoa(address));

		 }
	       }
		break;

	case TC_PACKET:

         tc = mes->olsr_tc;
         fprintf(fd, "Packet seq number: %d \n", ntohs(tc->tc_seqnum));
         fprintf(fd, "MPRS Sequence number: %d \n", ntohs(tc->tc_mprseqnum));
         fprintf(fd, "Hop Count:  %d \n", tc->tc_hopcnt);
	
         address.s_addr = tc->tc_origaddr;

	 fprintf(fd, "Originator address %s\n", inet_ntoa(address));

          
           for (maddr = tc->tc_mprsaddr; (char *)maddr < (char *)mes 
                                  + ntohs(mes->olsr_nextmsg); 
                maddr++){

	    address.s_addr = *maddr;

	    fprintf(fd, "MPRS address %s\n", inet_ntoa(address));

	   }
 
        break;

	case OLSRD_TRACEON:
		fprintf(fd, "\tfile=%*s\n", size, m->olsr_tracefile);
		break;

	case OLSRD_TRACEOFF:
		break;

	}
	}
	fflush(fd);
	if (ferror(fd))
		traceoff();
}
